@@ -21,18 +21,47 @@ typedef struct {
 
 
 static void
-pg_copycoder_mark( t_pg_copycoder *this )
+pg_copycoder_mark( void *_this )
 {
-	pg_coder_mark(&this->comp);
-	rb_gc_mark(this->typemap);
-	rb_gc_mark(this->null_string);
+	t_pg_copycoder *this = (t_pg_copycoder *)_this;
+	rb_gc_mark_movable(this->typemap);
+	rb_gc_mark_movable(this->null_string);
 }
 
+static size_t
+pg_copycoder_memsize( const void *_this )
+{
+	const t_pg_copycoder *this = (const t_pg_copycoder *)_this;
+	return sizeof(*this);
+}
+
+static void
+pg_copycoder_compact( void *_this )
+{
+	t_pg_copycoder *this = (t_pg_copycoder *)_this;
+	pg_coder_compact(&this->comp);
+	pg_gc_location(this->typemap);
+	pg_gc_location(this->null_string);
+}
+
+static const rb_data_type_t pg_copycoder_type = {
+	"PG::CopyCoder",
+	{
+		pg_copycoder_mark,
+		RUBY_TYPED_DEFAULT_FREE,
+		pg_copycoder_memsize,
+		pg_compact_callback(pg_copycoder_compact),
+	},
+	&pg_coder_type,
+	0,
+	RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE
 pg_copycoder_encoder_allocate( VALUE klass )
 {
 	t_pg_copycoder *this;
-	VALUE self = Data_Make_Struct( klass, t_pg_copycoder, pg_copycoder_mark, -1, this );
+	VALUE self = TypedData_Make_Struct( klass, t_pg_copycoder, &pg_copycoder_type, this );
 	pg_coder_init_encoder( self );
 	this->typemap = pg_typemap_all_strings;
 	this->delimiter = '\t';
@@ -44,7 +73,7 @@ static VALUE
 pg_copycoder_decoder_allocate( VALUE klass )
 {
 	t_pg_copycoder *this;
-	VALUE self = Data_Make_Struct( klass, t_pg_copycoder, pg_copycoder_mark, -1, this );
+	VALUE self = TypedData_Make_Struct( klass, t_pg_copycoder, &pg_copycoder_type, this );
 	pg_coder_init_decoder( self );
 	this->typemap = pg_typemap_all_strings;
 	this->delimiter = '\t';
@@ -63,7 +92,7 @@ pg_copycoder_decoder_allocate( VALUE klass )
 static VALUE
 pg_copycoder_delimiter_set(VALUE self, VALUE delimiter)
 {
-	t_pg_copycoder *this = DATA_PTR(self);
+	t_pg_copycoder *this = RTYPEDDATA_DATA(self);
 	StringValue(delimiter);
 	if(RSTRING_LEN(delimiter) != 1)
 		rb_raise( rb_eArgError, "delimiter size must be one byte");
@@ -80,7 +109,7 @@ pg_copycoder_delimiter_set(VALUE self, VALUE delimiter)
 static VALUE
 pg_copycoder_delimiter_get(VALUE self)
 {
-	t_pg_copycoder *this = DATA_PTR(self);
+	t_pg_copycoder *this = RTYPEDDATA_DATA(self);
 	return rb_str_new(&this->delimiter, 1);
 }
 
@@ -93,7 +122,7 @@ pg_copycoder_delimiter_get(VALUE self)
 static VALUE
 pg_copycoder_null_string_set(VALUE self, VALUE null_string)
 {
-	t_pg_copycoder *this = DATA_PTR(self);
+	t_pg_copycoder *this = RTYPEDDATA_DATA(self);
 	StringValue(null_string);
 	this->null_string = null_string;
 	return null_string;
@@ -105,7 +134,7 @@ pg_copycoder_null_string_set(VALUE self, VALUE null_string)
 static VALUE
 pg_copycoder_null_string_get(VALUE self)
 {
-	t_pg_copycoder *this = DATA_PTR(self);
+	t_pg_copycoder *this = RTYPEDDATA_DATA(self);
 	return this->null_string;
 }
 
@@ -123,7 +152,7 @@ pg_copycoder_null_string_get(VALUE self)
 static VALUE
 pg_copycoder_type_map_set(VALUE self, VALUE type_map)
 {
-	t_pg_copycoder *this = DATA_PTR( self );
+	t_pg_copycoder *this = RTYPEDDATA_DATA( self );
 
 	if ( !rb_obj_is_kind_of(type_map, rb_cTypeMap) ){
 		rb_raise( rb_eTypeError, "wrong elements type %s (expected some kind of PG::TypeMap)",
@@ -143,7 +172,7 @@ pg_copycoder_type_map_set(VALUE self, VALUE type_map)
 static VALUE
 pg_copycoder_type_map_get(VALUE self)
 {
-	t_pg_copycoder *this = DATA_PTR( self );
+	t_pg_copycoder *this = RTYPEDDATA_DATA( self );
 
 	return this->typemap;
 }
@@ -188,7 +217,7 @@ pg_text_enc_copy_row(t_pg_coder *conv, VALUE value, char *out, VALUE *intermedia
 	char *current_out;
 	char *end_capa_ptr;
 
-	p_typemap = DATA_PTR( this->typemap );
+	p_typemap = RTYPEDDATA_DATA( this->typemap );
 	p_typemap->funcs.fit_to_query( this->typemap, value );
 
 	/* Allocate a new string with embedded capacity and realloc exponential when needed. */
@@ -225,7 +254,7 @@ pg_text_enc_copy_row(t_pg_coder *conv, VALUE value, char *out, VALUE *intermedia
 
 				if( strlen == -1 ){
 					/* we can directly use String value in subint */
-					strlen = RSTRING_LEN(subint);
+					strlen = RSTRING_LENINT(subint);
 
 					/* size of string assuming the worst case, that every character must be escaped. */
 					PG_RB_STR_ENSURE_CAPA( *intermediate, strlen * 2, current_out, end_capa_ptr );
@@ -376,7 +405,7 @@ pg_text_dec_copy_row(t_pg_coder *conv, const char *input_line, int len, int _tup
 	char *end_capa_ptr;
 	t_typemap *p_typemap;
 
-	p_typemap = DATA_PTR( this->typemap );
+	p_typemap = RTYPEDDATA_DATA( this->typemap );
 	expected_fields = p_typemap->funcs.fit_to_copy_get( this->typemap );
 
 	/* The received input string will probably have this->nfields fields. */
@@ -397,7 +426,7 @@ pg_text_dec_copy_row(t_pg_coder *conv, const char *input_line, int len, int _tup
 		int found_delim = 0;
 		const char *start_ptr;
 		const char *end_ptr;
-		int input_len;
+		long input_len;
 
 		/* Remember start of field on input side */
 		start_ptr = cur_ptr;
