@@ -10,7 +10,6 @@
 #include "pg.h"
 
 static VALUE rb_cTypeMapByClass;
-static ID s_id_ancestors;
 
 typedef struct {
 	t_typemap typemap;
@@ -48,7 +47,7 @@ pg_tmbk_lookup_klass(t_tmbk *this, VALUE klass, VALUE param_value)
 
 		if( NIL_P(obj) ){
 			int i;
-			VALUE ancestors = rb_funcall( klass, s_id_ancestors, 0 );
+			VALUE ancestors = rb_mod_ancestors( klass );
 
 			Check_Type( ancestors, T_ARRAY );
 			/* Don't look at the first element, it's expected to equal klass. */
@@ -63,7 +62,7 @@ pg_tmbk_lookup_klass(t_tmbk *this, VALUE klass, VALUE param_value)
 		if(NIL_P(obj)){
 			p_coder = NULL;
 		}else if(rb_obj_is_kind_of(obj, rb_cPG_Coder)){
-			Data_Get_Struct(obj, t_pg_coder, p_coder);
+			TypedData_Get_Struct(obj, t_pg_coder, &pg_coder_type, p_coder);
 		}else{
 			if( RB_TYPE_P(obj, T_SYMBOL) ){
 				/* A Symbol: Call the method with this name. */
@@ -75,11 +74,9 @@ pg_tmbk_lookup_klass(t_tmbk *this, VALUE klass, VALUE param_value)
 
 			if( NIL_P(obj) ){
 				p_coder = NULL;
-			}else if( rb_obj_is_kind_of(obj, rb_cPG_Coder) ){
-				Data_Get_Struct(obj, t_pg_coder, p_coder);
 			}else{
-				rb_raise(rb_eTypeError, "argument has invalid type %s (should be nil or some kind of PG::Coder)",
-							rb_obj_classname( obj ));
+				/* Check retrieved coder type */
+				TypedData_Get_Struct(obj, t_pg_coder, &pg_coder_type, p_coder);
 			}
 
 			/* We can not cache coders retrieved by ruby code, because we can not anticipate
@@ -104,7 +101,7 @@ pg_tmbk_typecast_query_param( t_typemap *p_typemap, VALUE param_value, int field
   p_coder = pg_tmbk_lookup_klass( this, rb_obj_class(param_value), param_value );
 
 	if( !p_coder ){
-		t_typemap *default_tm = DATA_PTR( this->typemap.default_typemap );
+		t_typemap *default_tm = RTYPEDDATA_DATA( this->typemap.default_typemap );
 		return default_tm->funcs.typecast_query_param( default_tm, param_value, field );
 	}
 
@@ -114,32 +111,62 @@ pg_tmbk_typecast_query_param( t_typemap *p_typemap, VALUE param_value, int field
 static VALUE
 pg_tmbk_fit_to_query( VALUE self, VALUE params )
 {
-	t_tmbk *this = (t_tmbk *)DATA_PTR(self);
+	t_tmbk *this = (t_tmbk *)RTYPEDDATA_DATA(self);
 	/* Nothing to check at this typemap, but ensure that the default type map fits. */
-	t_typemap *default_tm = DATA_PTR( this->typemap.default_typemap );
+	t_typemap *default_tm = RTYPEDDATA_DATA( this->typemap.default_typemap );
 	default_tm->funcs.fit_to_query( this->typemap.default_typemap, params );
 	return self;
 }
 
 static void
-pg_tmbk_mark( t_tmbk *this )
+pg_tmbk_mark( void *_this )
 {
-	rb_gc_mark(this->typemap.default_typemap);
-	rb_gc_mark(this->klass_to_coder);
-	rb_gc_mark(this->self);
-	/* Clear the cache, to be safe from changes of klass VALUE by GC.compact.
-	 * TODO: Move cache clearing to compactation callback provided by Ruby-2.7+.
-	 */
+	t_tmbk *this = (t_tmbk *)_this;
+	pg_typemap_mark(&this->typemap);
+	rb_gc_mark_movable(this->klass_to_coder);
+	rb_gc_mark_movable(this->self);
+}
+
+static size_t
+pg_tmbk_memsize( const void *_this )
+{
+	const t_tmbk *this = (const t_tmbk *)_this;
+	return sizeof(*this);
+}
+
+static void
+pg_tmbk_compact(void *ptr)
+{
+	t_tmbk *this = (t_tmbk *)ptr;
+
+	pg_typemap_compact(&this->typemap);
+	pg_gc_location(this->klass_to_coder);
+	pg_gc_location(this->self);
+
+	/* Clear the cache, to be safe from changes of klass VALUE by GC.compact. */
 	memset(&this->cache_row, 0, sizeof(this->cache_row));
 }
 
+static const rb_data_type_t pg_tmbk_type = {
+	"PG::TypeMapByClass",
+	{
+		pg_tmbk_mark,
+		RUBY_TYPED_DEFAULT_FREE,
+		pg_tmbk_memsize,
+		pg_compact_callback(pg_tmbk_compact),
+	},
+	&pg_typemap_type,
+	0,
+	RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
 static VALUE
 pg_tmbk_s_allocate( VALUE klass )
 {
 	t_tmbk *this;
 	VALUE self;
 
-	self = Data_Make_Struct( klass, t_tmbk, pg_tmbk_mark, -1, this );
+	self = TypedData_Make_Struct( klass, t_tmbk, &pg_tmbk_type, this );
 	this->typemap.funcs.fit_to_result = pg_typemap_fit_to_result;
 	this->typemap.funcs.fit_to_query = pg_tmbk_fit_to_query;
 	this->typemap.funcs.fit_to_copy_get = pg_typemap_fit_to_copy_get;
@@ -153,7 +180,7 @@ pg_tmbk_s_allocate( VALUE klass )
 	this->self = self;
 	this->klass_to_coder = rb_hash_new();
 
-	/* The cache is properly initialized by Data_Make_Struct(). */
+	/* The cache is properly initialized by TypedData_Make_Struct(). */
 
 	return self;
 }
@@ -176,7 +203,7 @@ pg_tmbk_s_allocate( VALUE klass )
 static VALUE
 pg_tmbk_aset( VALUE self, VALUE klass, VALUE coder )
 {
-	t_tmbk *this = DATA_PTR( self );
+	t_tmbk *this = RTYPEDDATA_DATA( self );
 
 	if(NIL_P(coder)){
 		rb_hash_delete( this->klass_to_coder, klass );
@@ -200,7 +227,7 @@ pg_tmbk_aset( VALUE self, VALUE klass, VALUE coder )
 static VALUE
 pg_tmbk_aref( VALUE self, VALUE klass )
 {
-	t_tmbk *this = DATA_PTR( self );
+	t_tmbk *this = RTYPEDDATA_DATA( self );
 
 	return rb_hash_lookup(this->klass_to_coder, klass);
 }
@@ -214,7 +241,7 @@ pg_tmbk_aref( VALUE self, VALUE klass )
 static VALUE
 pg_tmbk_coders( VALUE self )
 {
-	t_tmbk *this = DATA_PTR( self );
+	t_tmbk *this = RTYPEDDATA_DATA( self );
 
 	return rb_obj_freeze(rb_hash_dup(this->klass_to_coder));
 }
@@ -222,8 +249,6 @@ pg_tmbk_coders( VALUE self )
 void
 init_pg_type_map_by_class()
 {
-	s_id_ancestors = rb_intern("ancestors");
-
 	/*
 	 * Document-class: PG::TypeMapByClass < PG::TypeMap
 	 *
